
import 'dart:convert';

import 'package:api_package/src/build_response.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:shelf/shelf.dart';

part 'ApiError.g.dart';

@JsonSerializable()
class ApiError {
  final String error;
  final String message;
  @JsonKey(ignore: true)
  final int statusCode;

  const ApiError({
    required this.error,
    required this.message,
    this.statusCode = 400,
  });

  static ApiError wrongParams(String paramName) => ApiError(
    error: "wrong_params",
    message: 'Invalid parameter ($paramName).',
  );

  static const ApiError noRoute = ApiError(
    error: "no_route",
    message: "Route not found.",
    statusCode: 404,
  );

  static const ApiError tokenInvalid = ApiError(
    error: "token_invalid",
    message: "The received token is invalid.",
    statusCode: 401,
  );

  static const ApiError tokenMissing = ApiError(
    error: "token_missing",
    message: "No token submitted in authentication header.",
    statusCode: 401,
  );

  static const ApiError accessDenied = ApiError(
    error: "access_denied",
    message: "You are not allowed to access the requested resource.",
    statusCode: 403,
  );

  static const ApiError notFound = ApiError(
    error: "not_found",
    message: "The requested resource does not exist.",
    statusCode: 404,
  );

  static const ApiError noData = ApiError(
    error: "no_data",
    message: "No data available.",
  );

  static const ApiError internalServerError = ApiError(
    statusCode: 500,
    error: "internal_server_error",
    message: "The server encountered an unexpected condition and can't fulfill the request.",
  );

  static const ApiError notImplemented = ApiError(
    statusCode: 501,
    error: "not_implemented",
    message: "Not implemented yet.",
  );

  Response build() => buildResponse(
    statusCode: statusCode,
    body: jsonEncode(this.toJson()),
    cache: null,
  );

  factory ApiError.fromJson(Map<String, dynamic> json) => _$ApiErrorFromJson(json);
  Map<String, dynamic> toJson() => _$ApiErrorToJson(this);
}

typedef Future<Response> ApiHandler();
int id = 0;
Future<Response> errorHandler(Request request, ApiHandler handler) async {
  try {
    Map<String, dynamic> queryParameters = request.requestedUri.queryParameters;
    Uri uri = Uri(
      path: request.requestedUri.path,
      queryParameters: queryParameters.isEmpty ? null : queryParameters,
    );
    print("${++id} | ${DateTime.now()} | ${request.method} ${Uri.decodeFull(uri.toString())}");
    return await handler();
  } on ApiError catch (error) {
    print("$id | ${DateTime.now()} | ${error.statusCode} | ${error.error} | ${error.message}");
    return error.build();
  } catch (error, stacktrace) {
    print("$id | ${DateTime.now()} | $error");
    print(stacktrace);
    return ApiError.internalServerError.build();
  }
}
