import 'dart:io';

import 'package:path/path.dart';
import 'package:postgres/postgres.dart';

import 'Env.dart';

/// Handles the connection with the pg database
class Database {
  static bool _init = false;
  PostgreSQLConnection _connection;

  Database._(this._connection);

  static Future<Database> get createNew async {
    if (!_init)
      throw Exception("Database.init() has not been called.");
    var connection = PostgreSQLConnection(
      Env.get.pgHost,
      Env.get.pgPort,
      Env.get.pgDatabase,
      username: Env.get.pgUser,
      password: Env.get.pgPassword,
    );
    await connection.open();
    return Database._(connection);
  }

  static Future<void> init() async {
    if (_init)
      throw Exception("Database.init() has already been called.");
	  _init = true;
    _DatabaseStructure.upgradeIfNeeded();
  }

  Future<List<Map<String, Map<String, dynamic>>>> queryMapped(String sql, {
    Map<String, dynamic>? values = const {}
  }) => _connection.mappedResultsQuery(sql, substitutionValues: values);

  Future<List<dynamic>> query(String sql, {Map<String, dynamic>? values = const {}}) =>
      _connection.query(sql, substitutionValues: values);

  @deprecated
  Future<void> patchSqlFile(File file) async {
    String content = await file.readAsString();
    List<String> operations = content.split(";");
    operations.removeLast();
    for (String operation in operations) {
      await queryMapped("$operation;");
    }
  }
}

class _DatabaseStructure {
  final Directory _sqlStructures = Directory(joinAll([
    Directory.current.path, "res", "sql", "structure",
  ]));
  final Directory _sqlMockData = Directory(joinAll([
    Directory.current.path, "res", "sql", "mock_data",
  ]));

  final Database db;

  _DatabaseStructure._(this.db) {
    if (!_sqlStructures.existsSync())
      throw Exception("Directory not found: ${_sqlStructures.path}");
    if (!_sqlMockData.existsSync())
      throw Exception("Directory not found: ${_sqlMockData.path}");
  }

  static Future<void> upgradeIfNeeded() async {
    var instance = _DatabaseStructure._(await Database.createNew);
    String? currentDbVersion = await instance.getDatabaseVersion();
    List<FileSystemEntity> databaseVersions = instance._sqlStructures.listSync();
    if (currentDbVersion != null) {
      // remove sql files that are not needed from list
      for (FileSystemEntity entity in databaseVersions) {
        databaseVersions.remove(entity);
        if (basename(entity.path) == "$currentDbVersion.sql") {
          break;
        }
      }
    }
    while (databaseVersions.isNotEmpty) {
      String path = databaseVersions.first.path;
      String fileName = basename(path);
      String newDbVersion = fileName.replaceFirst(".sql", "");
      print(
          "Using '$fileName' to migrate database from ${currentDbVersion == null ? "scratch" : "version $currentDbVersion"} to $newDbVersion.");
      await instance.patchDatabaseVersion(newDbVersion);
      currentDbVersion = newDbVersion;
      if (Env.debug) {
        // debug mode, insert mock data for testing
        var file = File(join(instance._sqlMockData.path, "$currentDbVersion.sql"));
        if (await file.exists()) {
          print("Inserting mock data for version $currentDbVersion.");
          await instance.db.patchSqlFile(file);
        }
      }
      databaseVersions.removeAt(0);
    }
  }

  Future<void> patchDatabaseVersion(String version) async {
    print("Updating database to version $version");
    db.patchSqlFile(File(join(_sqlStructures.path, "$version.sql")));
    await db.query("INSERT INTO `version`(`version`) VALUES ('$version');");
  }

  Future<bool> structureExists() async {
    List results = await db.query("SHOW TABLES LIKE 'version';");
    return results.isNotEmpty;
  }

  Future<String?> getDatabaseVersion() async {
    if (!await structureExists()) return null;
    var results = await db.queryMapped("SELECT `version` FROM `version` ORDER BY id DESC LIMIT 1;");
    return results.first["version"]!["version"];
  }
}